mod data;
mod generators;
mod templates;
mod parser;

use std::collections::HashMap;
use std::fmt::Write;
use std::fs::File;
use std::io::Read;
use clap::Parser;

#[derive(Debug, clap::Parser)]
struct Args {
    pub file: std::path::PathBuf,
    #[arg(short='n', long="name", help="base name for output files", required=true)]
    pub rpc_name: std::path::PathBuf,
    #[command(flatten)]
    pub generators: GenArgs
}

#[derive(Debug, clap::Args)]
#[group(required = true, multiple = true)]
struct GenArgs {
    #[arg(value_enum, short='c', long="client")]
    pub clients: Vec<generators::ClientGenerators>,
    #[arg(value_enum, short='s', long="server")]
    pub servers: Vec<generators::ServerGenerators>
}

static SOURCE_FILE: once_cell::sync::OnceCell<String> = once_cell::sync::OnceCell::new();
static SOURCE: once_cell::sync::OnceCell<String> = once_cell::sync::OnceCell::new();

fn has_generic(ty: &mut data::Types) -> Option<&mut data::Types> {
    use data::Types;
    match ty {
        Types::String | Types::Bool |
        Types::F32 | Types::F64 |
        Types::I8 | Types::I16 | Types::I32 | Types::I64 |
        Types::U8 | Types::U16 | Types::U32 | Types::U64 => None,
        Types::Array(inner) => has_generic(inner),
        Types::Optional(inner) => has_generic(inner),
        Types::Named(_) => None,
        Types::Generic(_, _, _) => Some(ty)
    }
}

pub fn gen_generic_name(ty: &data::Types) -> String {
    use data::Types;
    match &ty {
        Types::String => "string".into(),
        Types::Bool => "bool".into(),
        Types::F32 => "f32".into(),
        Types::F64 => "f64".into(),
        Types::I8 => "i8".into(),
        Types::I16 => "i16".into(),
        Types::I32 => "i32".into(),
        Types::I64 => "i64".into(),
        Types::U8 => "u8".into(),
        Types::U16 => "u16".into(),
        Types::U32 => "u32".into(),
        Types::U64 => "u64".into(),
        Types::Named(name) => name.clone(),
        Types::Optional(inner) => format!("maybe_{}", gen_generic_name(inner)),
        Types::Array(inner) => format!("array_{}", gen_generic_name(inner)),
        Types::Generic(_, _, span) => emit_error(Some((span.clone(), "Nested custom generics are not allowed")))
    }
}
fn replace_generics<'a>(it: impl Iterator<Item=&'a mut data::Types>, generics: &HashMap<String, data::StructTy>, new_structs: &mut HashMap<String, (data::StructTy, data::Types)>) {
    for ty in it.filter_map(|ty| has_generic(ty)) {
        let (name, inner, span) = match ty {
            data::Types::Generic(name, inner, span) =>
                (name.clone(), inner.clone(), span.clone()),
            _ => unreachable!()
        };
        let generic = match generics.get(&name) {
            Some(v) => v,
            None => emit_error(Some((span, "Type does not exists")))
        };
        let new_name = format!("{}_{}", name, gen_generic_name(&inner));
        *ty = data::Types::Named(new_name.clone());
        new_structs.insert(new_name, (generic.clone(), *inner));
    }
}

fn replace_generic_type(ty: &mut data::Types, generic_name: &String, replacement: &data::Types) {
    use data::Types;
    match ty {
        Types::Named(name) => if name == generic_name { *ty = replacement.clone(); } else { unreachable!() },
        Types::Optional(inner) => replace_generic_type(inner, generic_name, replacement),
        Types::Array(inner) => replace_generic_type(inner, generic_name, replacement),
        _ => unreachable!()
    }
}

fn resolve_generics(rpc: &mut data::RPC) {
    let mut generics: HashMap<String, data::StructTy> = HashMap::new();
    rpc.structs.retain(|s| if s.generic_name.is_none() { true } else {
        generics.insert(s.name.clone(), s.clone());
        false
    });

    let mut new_structs = HashMap::new();
    for s in &mut rpc.structs {
        replace_generics(s.fields.iter_mut().map(|f| &mut f.ty), &generics, &mut new_structs);
    }
    for m in &mut rpc.services.iter_mut().map(|s| s.methods.iter_mut()).flatten() {
        replace_generics(m.args.iter_mut().map(|f| &mut f.ty).chain(&mut m.ret), &generics, &mut new_structs);
    }

    for (name, (mut generic, inner)) in new_structs {
        generic.name = name;
        let generic_name = generic.generic_name.take().unwrap();
        for mut field in generic.generic_fields {
            replace_generic_type(&mut field.ty, &generic_name, &inner);
            generic.fields.push(field);
        }
        generic.generic_fields = vec![];
        rpc.structs.push(generic);
    }
}

fn main() {
    let args = Args::parse();

    let mut file = File::open(args.file.clone()).unwrap();
    let mut content = String::new();
    file.read_to_string(&mut content).unwrap();

    SOURCE_FILE.set(args.file.to_string_lossy().to_string()).unwrap();
    SOURCE.set(content).unwrap();

    let ast = match syn::parse_file(SOURCE.get().unwrap()) {
        Ok(v) => v,
        Err(e) => emit_error(e.into_iter()
            .map(|e| {
                let span = e.span();
                let mut msg = String::new();
                write!(msg, "{e}").unwrap();
                (span, msg)
            })
        )
    };

    let mut rpc = parser::parse_file(&ast);

    resolve_generics(&mut rpc);

    for gen in &args.generators.clients {
        gen.generate(&args.rpc_name, &rpc);
    }

    for gen in &args.generators.servers {
        gen.generate(&args.rpc_name, &rpc);
    }
}

pub fn emit_error(errors: impl IntoIterator<Item=(proc_macro2::Span, impl Into<String>)>) -> ! {
    use codespan_reporting::{
        diagnostic::{Diagnostic, Label},
        files::{SimpleFiles, Files},
        term::{
            self,
            termcolor::{ColorChoice, StandardStream},
        },
    };
    let mut files = SimpleFiles::new();
    let config = term::Config::default();
    let writer = StandardStream::stderr(ColorChoice::Auto);
    let file_id = files.add(SOURCE_FILE.get().unwrap().clone(), SOURCE.get().unwrap().clone());

    let file = files.get(file_id).unwrap();

    let errors = errors.into_iter().map(|(span, msg)| {
        let start = span.start();
        let start: usize = file.line_range((), start.line-1).unwrap().start + start.column;
        let end = span.end();
        let end: usize = file.line_range((), end.line-1).unwrap().start + end.column;
        Label::primary(file_id, start..end).with_message(msg)
    });

    let diagnostic = Diagnostic::error()
        .with_labels(errors.collect());

    term::emit(&mut writer.lock(), &config, &files, &diagnostic).expect("cannot write error");
    std::process::exit(1);
}
